"""Provide support to Lexicon for AWS Route 53 DNS changes."""
import hashlib
import logging
import re

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

try:
    import boto3  # type: ignore
    import botocore  # type: ignore
except ImportError:
    pass

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = [re.compile(r"^awsdns-\d+\.\w+$")]


def provider_parser(subparser):
    """Specify arguments for AWS Route 53 Lexicon Provider."""
    subparser.add_argument(
        "--auth-access-key", help="specify ACCESS_KEY for authentication"
    )
    subparser.add_argument(
        "--auth-access-secret", help="specify ACCESS_SECRET for authentication"
    )
    subparser.add_argument(
        "--private-zone",
        help=(
            "indicates what kind of hosted zone to use. If true, use "
            "only private zones. If false, use only public zones"
        ),
    )

    # TODO: these are only required for testing, we should figure out
    # a way to remove them & update the integration tests
    # to dynamically populate the auth credentials that are required.
    subparser.add_argument(
        "--auth-username",
        help="alternative way to specify the ACCESS_KEY for authentication",
    )
    subparser.add_argument(
        "--auth-token",
        help="alternative way to specify the ACCESS_SECRET for authentication",
    )


class RecordSetPaginator(object):
    """Paginate through complete list of record sets."""

    def __init__(self, r53_client, hosted_zone_id, max_items=None):
        """Initialize paginator."""
        self.r53_client = r53_client
        self.hosted_zone_id = hosted_zone_id
        self.max_items = max_items

    def get_record_sets(self, **kwargs):
        """Retrieve a page from API."""
        return self.r53_client.list_resource_record_sets(**kwargs)

    def get_base_kwargs(self):
        """Get base kwargs for API call."""
        kwargs = {"HostedZoneId": self.hosted_zone_id}
        if self.max_items is not None:
            kwargs.update({"MaxItems": str(self.max_items)})
        return kwargs

    def all_record_sets(self):
        """Generator to loop through current record set.

        Call next page if it exists.
        """
        is_truncated = True
        start_record_name = None
        start_record_type = None
        kwargs = self.get_base_kwargs()
        while is_truncated:
            if start_record_name is not None:
                kwargs.update(
                    {
                        "StartRecordName": start_record_name,
                        "StartRecordType": start_record_type,
                    }
                )
            result = self.get_record_sets(**kwargs)
            for record_set in result.get("ResourceRecordSets", []):
                yield record_set

            is_truncated = result.get("IsTruncated", False)

            start_record_name = result.get("NextRecordName", None)
            start_record_type = result.get("NextRecordType", None)


class Provider(BaseProvider):
    """Provide AWS Route 53 implementation of Lexicon Provider interface."""

    def __init__(self, config):
        """Initialize AWS Route 53 DNS provider."""
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.private_zone = self._get_provider_option("private_zone")
        # instantiate the client
        self.r53_client = boto3.client(
            "route53",
            aws_access_key_id=self._get_provider_option("auth_access_key")
            or self._get_provider_option("auth_username"),
            aws_secret_access_key=self._get_provider_option("auth_access_secret")
            or self._get_provider_option("auth_token"),
        )

    def filter_zone(self, data):
        """Check if a zone is private"""
        if self.private_zone is not None:
            if data["Config"]["PrivateZone"] != self.str2bool(self.private_zone):
                return False

        if data["Name"] != f"{self.domain}.":
            return False

        return True

    @staticmethod
    def str2bool(input_string):
        """Convert a string to boolean"""
        return input_string.lower() in ("true", "yes")

    def _authenticate(self):
        """Determine the hosted zone id for the domain."""
        try:
            is_truncated = True
            next_dns_name = None
            next_hz_id = None
            hosted_zones = []

            while is_truncated:
                if not next_dns_name:
                    new_zones = self.r53_client.list_hosted_zones_by_name()
                else:
                    new_zones = self.r53_client.list_hosted_zones_by_name(
                        DNSName=next_dns_name, HostedZoneId=next_hz_id
                    )

                hosted_zones.extend(new_zones.get("HostedZones"))
                is_truncated = new_zones.get("IsTruncated")

                if is_truncated:
                    next_dns_name = new_zones.get("NextDNSName")
                    next_hz_id = new_zones.get("NextHostedZoneId")

            hosted_zone = next(hz for hz in hosted_zones if self.filter_zone(hz))
            self.domain_id = hosted_zone.get("Id")
        except StopIteration:
            raise AuthenticationError("No domain found")

    def _change_record_sets(self, action, rtype, name, content):
        ttl = self._get_lexicon_option("ttl")
        resource_records = []
        if isinstance(content, list):
            for i in content:
                value = f'"{i}"' if rtype in ["TXT", "SPF"] else i
                resource_records.append({"Value": value})
        else:
            value = f'"{content}"' if rtype in ["TXT", "SPF"] else content
            resource_records.append({"Value": value})
        try:
            self.r53_client.change_resource_record_sets(
                HostedZoneId=self.domain_id,
                ChangeBatch={
                    "Comment": f"{action} using lexicon Route 53 provider",
                    "Changes": [
                        {
                            "Action": action,
                            "ResourceRecordSet": {
                                "Name": self._fqdn_name(name),
                                "Type": rtype,
                                "TTL": ttl if ttl is not None else 300,
                                "ResourceRecords": resource_records,
                            },
                        }
                    ],
                },
            )
            return True
        except botocore.exceptions.ClientError as error:
            if "Duplicate Resource Record" in error.response["Error"]["Message"]:
                # Duplicate resource, that have been a noop. This is expected.
                return True
            LOGGER.error(str(error), exc_info=True)
            return False

    def _create_record(self, rtype, name, content):
        """Create a record in the hosted zone."""
        existing_records = self._list_record_sets(rtype, name)
        if existing_records:
            existing_record = existing_records[0]
            if isinstance(existing_records[0]["content"], list):
                return self._change_record_sets(
                    "UPSERT",
                    existing_record["type"],
                    existing_record["name"],
                    existing_record["content"] + [content],
                )
            return self._change_record_sets(
                "UPSERT", rtype, name, [existing_record["content"]] + [content]
            )
        return self._change_record_sets("CREATE", rtype, name, content)

    def _update_record(self, identifier=None, rtype=None, name=None, content=None):
        """Update a record from the hosted zone."""
        if identifier:
            records = [
                record
                for record in self._list_records()
                if identifier == _identifier(record)
            ]
            if not records:
                raise ValueError(f"No record found for identifier {identifier}")
            record = records[0]
            rtype = record["type"]
            name = record["name"]

        existing_records = self._list_record_sets(rtype, name)
        if not existing_records:
            raise ValueError("No matching record to update was found.")

        for existing_record in existing_records:
            if isinstance(existing_record["content"], list):
                # Multiple values in record.
                LOGGER.warning(
                    "Warning, multiple records found for given parameters, "
                    "only first entry will be updated: %s",
                    existing_record,
                )
                new_content = existing_record["content"].copy()
                new_content[0] = content
            else:
                new_content = content

            self._change_record_sets(
                "UPSERT", existing_record["type"], existing_record["name"], new_content
            )

        return True

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        """Delete a record from the hosted zone."""
        if identifier:
            matching_records = [
                record
                for record in self._list_records()
                if identifier == _identifier(record)
            ]
            if not matching_records:
                raise ValueError(f"No record found for identifier {identifier}")
            rtype = matching_records[0]["type"]
            name = matching_records[0]["name"]
            content = matching_records[0]["content"]

        existing_records = self._list_record_sets(rtype, name, content)
        if not existing_records:
            raise ValueError("No record found for the provided type, name and content")

        for existing_record in existing_records:
            if isinstance(existing_record["content"], list) and content is not None:
                # multiple values in record, just remove one value and only if it actually exist
                if content in existing_record["content"]:
                    existing_record["content"].remove(content)
                    self._change_record_sets(
                        "UPSERT",
                        existing_record["type"],
                        existing_record["name"],
                        existing_record["content"],
                    )
            else:
                # if only one record exist, or if content is not specified, remove whole record
                self._change_record_sets(
                    "DELETE",
                    existing_record["type"],
                    existing_record["name"],
                    existing_record["content"],
                )

        return True

    def _list_records(self, rtype=None, name=None, content=None):
        """List all records for the hosted zone."""
        records = self._list_record_sets(rtype, name, content)

        flatten_records = []
        for record in records:
            if isinstance(record["content"], list):
                for one_content in record["content"]:
                    flatten_record = record.copy()
                    flatten_record["content"] = one_content
                    flatten_record["id"] = _identifier(flatten_record)
                    flatten_records.append(flatten_record)
            else:
                record["id"] = _identifier(record)
                flatten_records.append(record)

        LOGGER.debug("list_records: %s", records)

        return flatten_records

    def _list_record_sets(self, rtype=None, name=None, content=None):
        records = []
        paginator = RecordSetPaginator(self.r53_client, self.domain_id)
        for record in paginator.all_record_sets():
            record_content = []
            if rtype is not None and record["Type"] != rtype:
                continue
            if name is not None and record["Name"] != self._fqdn_name(name):
                continue
            if record.get("AliasTarget", None) is not None:
                record_content = [record["AliasTarget"].get("DNSName", None)]
            if record.get("ResourceRecords", None) is not None:
                record_content = [
                    _format_content(record["Type"], value["Value"])
                    for value in record["ResourceRecords"]
                ]
            if content is not None and content not in record_content:
                continue

            LOGGER.debug("record: %s", record)
            records.append(
                {
                    "type": record["Type"],
                    "name": self._full_name(record["Name"]),
                    "ttl": record.get("TTL", None),
                    "content": record_content[0]
                    if len(record_content) == 1
                    else record_content,
                }
            )
        return records

    def _request(self, action="GET", url="/", data=None, query_params=None):
        # Helper _request is not used in Route53 provider
        pass


def _format_content(rtype, content):
    return content[1:-1] if rtype in ["TXT", "SPF"] else content


def _identifier(record):
    sha256 = hashlib.sha256()
    sha256.update(("type=" + record.get("type", "") + ",").encode("utf-8"))
    sha256.update(("name=" + record.get("name", "") + ",").encode("utf-8"))
    sha256.update(("data=" + record.get("data", "") + ",").encode("utf-8"))
    return sha256.hexdigest()[0:7]



"""Integration tests for Constellix"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Constellix does not currently have a sandbox and they enforce domain
# uniqueness across the service.  You'll need your own production credentials
# and a unique domain name if you want to run these tests natively.
class ConstellixProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for Constellix"""

    provider_name = "constellix"
    domain = "example.org"

    def _filter_headers(self):
        return ["x-cnsdns-apiKey", "x-cnsdns-hmac", "x-cnsdns-requestDate"]



"""Integration tests for DNSPark"""
from unittest import TestCase

import pytest

from lexicon.tests.providers.integration_tests import IntegrationTestsV1


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
# TODO: migrate to IntegrationTestsV2 and its extended test suite
class DnsParkProviderTests(TestCase, IntegrationTestsV1):
    """TestCase for DNSPark"""

    provider_name = "dnspark"
    domain = "capsulecd.com"

    def _filter_headers(self):
        return ["Authorization"]

    # TODO: enable the skipped tests
    @pytest.mark.skip(reason="new test, missing recording")
    def test_provider_when_calling_update_record_should_modify_record_name_specified(
        self,
    ):
        return

    @pytest.mark.skip(reason="new test, missing recording")
    def test_provider_when_calling_list_records_after_setting_ttl(self):
        return



"""Module provider for NSOne"""
import json
import logging

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["nsone.net"]


def provider_parser(subparser):
    """Configure provider parser for NSOne"""
    subparser.add_argument("--auth-token", help="specify token for authentication")


class Provider(BaseProvider):
    """Provider class for NSOne"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://api.nsone.net/v1"

    def _authenticate(self):

        payload = self._get(f"/zones/{self.domain}")

        if not payload["id"]:
            raise AuthenticationError("No domain found")

        self.domain_id = self.domain

    def _get_record_set(self, name, rtype):
        try:
            payload = self._get(f"/zones/{self.domain_id}/{name}/{rtype}")
        except requests.exceptions.HTTPError as error:
            if error.response.status_code == 404:
                return None
            raise

        return {
            "type": payload["type"],
            "name": payload["domain"],
            "ttl": payload["ttl"],
            "answers": payload["answers"],
        }

    # Create record. If record already exists with the same content, do nothing'
    def _create_record(self, rtype, name, content):
        name = self._full_name(name)
        existing_record_set = self._get_record_set(name, rtype)
        if existing_record_set:

            def _record_set_has_answer(record_set, content):
                for answer in record_set["answers"]:
                    if content in answer["answer"]:
                        return True
                return False

            if not _record_set_has_answer(existing_record_set, content):
                existing_record_set["answers"].append({"answer": [content]})
                self._post(
                    f"/zones/{self.domain_id}/{name}/{rtype}", existing_record_set
                )
        else:
            record = {
                "type": rtype,
                "domain": name,
                "ttl": self._get_lexicon_option("ttl"),
                "zone": self.domain_id,
                "answers": [{"answer": [content]}],
            }
            payload = {}
            try:
                payload = self._put(f"/zones/{self.domain_id}/{name}/{rtype}", record)
            except requests.exceptions.HTTPError as error:
                # http 400 is ok here, because the record probably already exists
                if error.response.status_code != 400:
                    raise

            LOGGER.debug("create_record: %s", "id" in payload)

        return True

    def _find_record(self, domain, _type=None):
        """search for a record on NS1 across zones. returns None if not found."""

        def _is_matching(record):
            """filter function for records"""

            if domain and record.get("domain", None) != domain:
                return False
            if _type and record.get("type", None) != _type:
                return False
            return True

        payload = self._get(f"/search?q={domain}&type=record")
        for record in payload:
            if _is_matching(record):
                match = record
                break
        else:
            # no such domain on ns1
            return None

        record = self._get(f"/zones/{match['zone']}/{match['domain']}/{match['type']}")
        if record.get("message", None):
            return None  # {"message":"record not found"}
        short_answers = [x["answer"][0] for x in record["answers"]]

        # ensure a compatibility level with self._list_records
        record["short_answers"] = short_answers
        return record

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        def _resolve_link(record, recurse=0):
            # https://ns1.com/articles/cname-alias-and-linked-records
            # - recursion is allowed
            # - link source and link target are always of the same rtype
            # - target can be anywhere on ns1, not necessarily self.domain_id.
            if record.get("link", None) is None:
                # not a linked record
                return record

            if recurse < 1:
                return None

            match = self._find_record(record["link"], _type=record["type"])
            if not match:
                return None

            return _resolve_link(match, recurse=recurse - 1)

        payload = self._get(f"/zones/{self.domain_id}")
        records = []
        for record in payload["records"]:

            if rtype and record["type"] != rtype:
                continue

            if name and record["domain"] != self._full_name(name):
                continue

            link_target = _resolve_link(record, recurse=3)

            if link_target and link_target.get("short_answers", None):
                # target found (could be the same as orig record)
                answers = link_target["short_answers"]
            else:
                # recursion limit reached. or unhandled record format.
                answers = []

            if content and content not in answers:
                continue

            for answer in answers:
                processed_record = {
                    "type": record["type"],
                    "name": record["domain"],
                    "ttl": record["ttl"],
                    "content": answer,
                    # this id is useless unless your doing record linking. Lets return the
                    # original record identifier.
                    "id": f"{self.domain_id}/{record['domain']}/{record['type']}",
                }
                records.append(processed_record)

        LOGGER.debug("list_records: %s", records)
        return records

    # Create or update a record.
    def _update_record(self, identifier, rtype=None, name=None, content=None):
        data = {}
        new_identifier = f"{self.domain_id}/{self._full_name(name)}/{rtype}"

        if new_identifier == identifier or (rtype is None and name is None):
            # the identifier hasnt changed, or type and name are both unspecified,
            # only update the content.
            data["answers"] = [{"answer": [content]}]
            self._post(f"/zones/{identifier}", data)

        else:
            # identifiers are different
            # get the old record, create a new one with updated data, delete the old record.
            old_record = self._get(f"/zones/{identifier}")
            self.create_record(
                rtype or old_record["type"],
                name or old_record["domain"],
                content or old_record["answers"][0]["answer"][0],
            )
            self.delete_record(identifier)

        LOGGER.debug("update_record: %s", True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        if not identifier:
            name = self._full_name(name)

            record_set = self._get_record_set(name, rtype)
            if record_set:
                record_set_new = {
                    "type": record_set["type"],
                    "name": record_set["name"],
                    "ttl": record_set["ttl"],
                    "answers": [],
                }

                if content:
                    for answer in record_set["answers"]:
                        if content not in answer["answer"]:
                            record_set_new["answers"].append(answer)

                if record_set_new["answers"]:
                    self._post(
                        f"/zones/{self.domain_id}/{name}/{rtype}", record_set_new
                    )
                else:
                    self._delete(f"/zones/{self.domain_id}/{name}/{rtype}")
        else:
            self._delete(f"/zones/{identifier}")

        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers
    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "X-NSONE-Key": self._get_provider_option("auth_token"),
        }
        default_auth = None

        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=json.dumps(data),
            headers=default_headers,
            auth=default_auth,
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response.json()



"""Module provider for DirectAdmin hosts"""
import logging
import warnings
from typing import List

import requests
from requests.auth import HTTPBasicAuth

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

# DirectAdmin is not tied to a specific domain, so there is nothing to specify here
NAMESERVER_DOMAINS: List[str] = []


def provider_parser(subparser):
    """Return the parser for this provider"""
    subparser.add_argument(
        "--auth-password",
        help="specify password for authentication (or login key for two-factor authentication)",
    )

    subparser.add_argument(
        "--auth-username", help="specify username for authentication"
    )

    subparser.add_argument("--endpoint", help="specify the DirectAdmin endpoint")


# See https://www.directadmin.com/features.php?id=504 for the specification of
# the URIs for the different operations
class Provider(BaseProvider):
    """Provider class for DirectAdmin"""

    def __init__(self, config):
        super(Provider, self).__init__(config)

        self.endpoint = self._get_provider_option("endpoint")
        if self.endpoint is None:
            raise Exception("Specify endpoint of DirectAdmin")

    def _authenticate(self):
        try:
            response = self._get("/CMD_API_SHOW_DOMAINS")
        except requests.exceptions.HTTPError as err:
            # A 401 error will be returned in case of incorrect or missing
            # credentials
            raise err

        # The response is a URL encoded array containing all available domains
        domains = [domain.split("=").pop() for domain in response.split("&")]
        try:
            self.domain_id = domains.index(self.domain)
        except ValueError:
            raise AuthenticationError(f"Domain {self.domain} not found")

    def _create_record(self, rtype, name, content):
        # Refuse to create duplicate records
        existing_records = self._list_records(rtype, name, content)
        if existing_records:
            return True

        query_params = {
            "action": "add",
            "json": "yes",
            "name": f"{self._full_name(name)}.",
            "type": rtype,
            "value": content,
        }

        if self._get_lexicon_option("ttl"):
            query_params["ttl"] = self._get_lexicon_option("ttl")

        try:
            response = self._get("/CMD_API_DNS_CONTROL", query_params)
        except requests.exceptions.HTTPError:
            response = {"success": "Create Failed"}

        LOGGER.debug("create_record: %s", response)

        return response["success"].lower().find("added") >= 0

    def _list_records(self, rtype=None, name=None, content=None):
        response = {"records": []}
        try:
            response = self._get("/CMD_API_DNS_CONTROL", {"json": "yes"})
        except requests.exceptions.HTTPError as err:
            warnings.warn(err.response.text)
            raise

        records = [
            self._parse_response_record(record) for record in response["records"]
        ]
        if rtype:
            records = [record for record in records if record["type"] == rtype]
        if name:
            cmp_name = self._full_name(name.lower())
            records = [record for record in records if record["name"] == cmp_name]
        if content:
            records = [record for record in records if record["content"] == content]

        LOGGER.debug("list_records: %s", records)

        return records

    def _parse_response_record(self, response_record):
        # Most fields in a record response match, except for content which is
        # value
        return {
            "content": response_record["value"],
            "id": response_record["combined"],
            "name": self._full_name(response_record["name"]),
            # 4 hours appears to be the default shown by DirectAdmin if no TTL
            # is set
            "ttl": response_record.get("ttl", 14400),
            "type": response_record["type"],
        }

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        # Editing a record is a combination of removing the old record and
        # adding a new one, but specifying 'edit' as the action while passing
        # all parameters necessary for both deletion and creation

        if not identifier:
            # The identifier of the original record to edit is necessary to be
            # able to create the appropriate delete payload
            original_records = self.list_records(rtype, name)
            if len(original_records) > 1:
                warnings.warn("Found multiple records to edited. Cannot continue...")
                return False

            if not original_records:
                warnings.warn("Found no records to edit. Cannot continue...")
                return False

            identifier = original_records[0]["id"]

        delete_key = self._determine_delete_key(identifier, rtype)

        query_params = {
            "action": "edit",
            "json": "yes",
            delete_key: identifier,
            "name": f"{self._full_name(name)}.",
            "type": rtype,
            "value": content,
        }

        try:
            response = self._get("/CMD_API_DNS_CONTROL", query_params)
        except requests.exceptions.HTTPError:
            response = {"success": "Update Failed"}

        LOGGER.debug("update_record: %s", response)

        return response["success"].lower().find("added") > 0

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        if not identifier:
            identifiers = [
                record["id"] for record in self._list_records(rtype, name, content)
            ]
        else:
            identifiers = [identifier]

        response = {"success": "noop"}
        for identifier_to_delete in identifiers:
            delete_key = self._determine_delete_key(identifier_to_delete, rtype)
            query_params = {
                "action": "select",
                "json": "yes",
                delete_key: identifier_to_delete,
            }

            try:
                response = self._get("/CMD_API_DNS_CONTROL", query_params)
            except requests.exceptions.HTTPError:
                response = {"success": "Delete Failed"}

            LOGGER.debug("delete_record: %s", response)

        return response["success"].lower().find("deleted") > 0

    def _determine_delete_key(self, identifier, rtype):
        # The indicator for the record that needs to be removed is determined
        # by the type of the record and its index within all records of that
        # type. There's an additional check on the name and value which still
        # need to match for the removal to actually occur
        if not rtype:
            # An rtype is necessary to create the delete_key. However, it may
            # not be specified in which case an effort needs to be made to
            # figure it out automatically. The necessary data can be recreated
            # from the identifier
            identifier_parts = identifier.split("&")
            name = identifier_parts[0].split("=").pop()
            content = identifier_parts[1].split("=").pop()

            records = self.list_records(None, name, content)
            rtype = records[0]["type"]

        existing_records = self.list_records(rtype)
        existing_record_index = 0
        for (index, record) in enumerate(existing_records):
            if record["id"] == identifier:
                existing_record_index = index

        return f"{rtype}recs{existing_record_index}".lower()

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}

        query_params["domain"] = self.domain

        response = requests.request(
            action,
            self.endpoint + url,
            auth=HTTPBasicAuth(
                self._get_provider_option("auth_username"),
                self._get_provider_option("auth_password"),
            ),
            params=query_params,
        )

        # If the request fails for any reason, throw an error
        response.raise_for_status()

        if "json" in response.headers["Content-Type"].lower():
            return response.json()

        return response.text



"""Integration tests for Joker.com provider"""
import re
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


class JokerProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for Joker.com"""

    provider_name = "joker"
    domain = "erdtfrau.de"

    def _filter_query_parameters(self):
        return [
            ("api-key", "DUMMY_API_KEY"),
            ("auth-sid", "DUMMY_AUTH_SID"),
        ]

    def _filter_response(self, response):
        body = response["body"]["string"].decode("utf-8")
        body = re.sub(r"Auth-Sid: \w+\n", "Auth-Sid: DUMMY_AUTH_SID\n", body)
        response["body"]["string"] = body.encode("utf-8")

        return response



"""
Lexicon UltraDNS Provider

Derived from PowerDNS provider

API Docs: https://doc.ultradns.com/md/httpapi/api_spec/

Implementation notes:
* The UltraDNS API does not assign a unique identifier to each record in the way
that Lexicon expects. We work around this by creating an ID based on the record
name, type and content, which when taken together are always unique
* The UltraDNS API has no notion of 'create a single record' or 'delete a single
record'. All operations are either 'replace the RRSet with this new set of records'
or 'delete all records for this name and type. Similarly, there is no notion of
'change the content of this record', because records are identified by their name,
type and content.
* The API is very picky about the format of values used when creating records:
** CNAMEs must be fully qualified
This is why the _clean_content and _unclean_content methods exist, to convert
back and forth between the format UltraDNS expects, and the format Lexicon uses
"""
import hashlib
import json
import logging

import requests

from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["ultradns.net"]


def provider_parser(subparser):
    """Configure provider parser for ultradns"""
    subparser.add_argument(
        "--auth-token",
        help="specify token for authentication;"
        + " if not set --auth-token, --auth-password are used",
    )
    subparser.add_argument(
        "--auth-username", help="specify username for authentication"
    )
    subparser.add_argument(
        "--auth-password", help="specify password for authentication"
    )


class Provider(BaseProvider):
    """Provider class for UltraDNS"""

    def __init__(self, config):
        super(Provider, self).__init__(config)

        self.api_key = self._get_provider_option("auth_token")

        assert (self.api_key is not None) or (
            self._get_provider_option("auth_username") is not None
            and self._get_provider_option("auth_password") is not None
        )

    def zone_data(self):
        """Get zone data"""
        data = self._get("/zones/" + self._ensure_dot(self.domain) + "/rrsets").json()
        for i, rrset in enumerate(data["rrSets"]):
            data["rrSets"][i]["rrtype"] = self._clean_rrtype(rrset["rrtype"])
        return data

    def _authenticate(self):
        if self.api_key is None:
            username = self._get_provider_option("auth_username")
            password = self._get_provider_option("auth_password")
            url = "https://restapi.ultradns.com/v2/authorization/token"
            data = {
                "grant_type": "password",
                "username": username,
                "password": password,
            }

            result = requests.post(url, data=data)
            result.raise_for_status()
            self.api_key = result.json()["accessToken"]

        assert self.api_key is not None
        self.domain_id = self.domain
        # Test auth works
        self.zone_data()

    def _make_identifier(self, rtype, name, content):
        sha256 = hashlib.sha256()
        sha256.update(("type=" + rtype + ",").encode("utf-8"))
        sha256.update(("name=" + name + ",").encode("utf-8"))
        sha256.update(("data=" + content + ",").encode("utf-8"))
        return sha256.hexdigest()[0:7]

    def _list_records(self, rtype=None, name=None, content=None):
        records = []
        for rrset in self.zone_data()["rrSets"]:
            if (
                name is None
                or self._fqdn_name(rrset["ownerName"]) == self._fqdn_name(name)
            ) and (rtype is None or rrset["rrtype"] == rtype):
                for record in rrset["rdata"]:
                    if content is None or record == self._clean_content(rtype, content):
                        records.append(
                            {
                                "type": rrset["rrtype"],
                                "name": self._full_name(rrset["ownerName"]),
                                "ttl": rrset["ttl"],
                                "content": self._unclean_content(
                                    rrset["rrtype"], record
                                ),
                                "id": self._make_identifier(
                                    rrset["rrtype"], rrset["ownerName"], record
                                ),
                            }
                        )
        LOGGER.debug("list_records: %s", records)
        return records

    def _clean_rrtype(self, rtype):
        """ UltraDNS returns records with types like 'MX (15)' """
        return rtype.split()[0]

    def _clean_content(self, rtype, content):
        if rtype == "CNAME" and not content.endswith("."):
            # Regularise non-FQDN CNAMEs - do not affect FQDNs or we break out of zone
            content = self._fqdn_name(content)
        return content

    def _unclean_content(self, rtype, content):
        if rtype == "CNAME" and not content.endswith("."):
            # Regularise non-FQDN CNAMEs - do not affect FQDNs or we break out of zone
            content = self._full_name(content)
        return content

    def _create_record(self, rtype, name, content):
        rname = self._fqdn_name(name)
        newcontent = self._clean_content(rtype, content)

        updated_data = {
            "ownerName": rname,
            "rrtype": rtype,
            "rdata": [],
            "ttl": self._get_lexicon_option("ttl") or 600,
        }

        updated_data["rdata"].append(newcontent)

        found = False
        for rrset in self.zone_data()["rrSets"]:
            if rrset["ownerName"] == rname and rrset["rrtype"] == rtype:
                updated_data["ttl"] = rrset["ttl"]
                found = True

                for record in rrset["rdata"]:
                    if record == newcontent:
                        return True  # Exactly the same record exists, just return
                    updated_data["rdata"].append(record)
                break

        if found:
            self._put(
                f"/zones/{self._ensure_dot(self.domain)}/rrsets/{rtype}/{rname}",
                data=updated_data,
            )
        else:
            self._post(
                f"/zones/{self._ensure_dot(self.domain)}/rrsets/{rtype}/{rname}",
                data=updated_data,
            )

        return True

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):

        data = self.zone_data()

        if identifier is not None:
            found = False
            for rrset in data["rrSets"]:
                for record in rrset["rdata"]:
                    ident = self._make_identifier(
                        rrset["rrtype"], rrset["ownerName"], record
                    )
                    if identifier == ident:
                        rtype = rrset["rrtype"]
                        name = self._full_name(rrset["ownerName"])
                        content = self._unclean_content(rrset["rrtype"], record)
                        found = True
                        break
                else:
                    continue
                break  # break out of the outer loop too
            if not found:
                return True  # No match means nothing to do

        LOGGER.debug("delete %s %s %s", rtype, name, content)
        if rtype is None or name is None:
            raise Exception("Must specify at least both rtype and name")

        rname = self._fqdn_name(name)

        found = False
        for rrset in data["rrSets"]:
            if (
                rrset["rrtype"] == rtype
                and self._fqdn_name(rrset["ownerName"]) == rname
            ):
                update_data = rrset
                found = True

                if content is None:
                    update_data["rdata"] = []
                else:
                    new_record_list = []
                    for record in update_data["rdata"]:
                        if self._clean_content(rrset["rrtype"], content) != record:
                            new_record_list.append(record)

                    update_data["rdata"] = new_record_list
                break

        if not found:
            return True  # Do nothing if the record did not exist

        request = {"rrSets": [update_data]}
        LOGGER.debug("request: %s", request)

        if update_data["rdata"]:
            self._put(
                f"/zones/{self._ensure_dot(self.domain)}/rrsets/{rtype}/{rname}",
                data=update_data,
            )
        else:
            self._delete(
                f"/zones/{self._ensure_dot(self.domain)}/rrsets/{rtype}/{rname}"
            )

        return True

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        self._delete_record(identifier, rtype, name, content)
        return self._create_record(rtype, name, content)

    def _patch(self, url="/", data=None, query_params=None):
        return self._request("PATCH", url, data=data, query_params=query_params)

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        response = requests.request(
            action,
            "https://restapi.ultradns.com/v2" + url,
            params=query_params,
            data=json.dumps(data),
            headers={
                "Authorization": "Bearer " + self.api_key,
                "Content-Type": "application/json",
                "Accept": "application/json",
            },
        )
        LOGGER.debug("response: %s", response.text)
        response.raise_for_status()
        return response

    @classmethod
    def _ensure_dot(cls, text):
        """
        This function makes sure a string contains a dot at the end
        """
        if text.endswith("."):
            return text
        return text + "."



"""Integration tests for Zilore"""
from unittest import TestCase

from lexicon.providers.zilore import Provider
from lexicon.tests.providers import integration_tests


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class ZiloreProviderTests(TestCase, integration_tests.IntegrationTestsV2):
    """TestCase for Zilore"""

    Provider = Provider
    provider_name = "zilore"
    domain = "full4ir.tk"

    def _filter_headers(self):
        return ["X-Auth-Key"]

    # We override this test because Zilore refuses to create a A record with '127.0.0.1' value.
    @integration_tests._vcr_integration_test
    def test_provider_when_calling_create_record_for_A_with_valid_name_and_content(
        self,
    ):
        provider = self._construct_authenticated_provider()
        assert provider.create_record("A", "localhost", "1.1.1.1")



""" Ensure that stdout corresponds to the given reference output """
import json

from lexicon import cli

DATA = [
    {
        "id": "fake-id",
        "type": "TXT",
        "name": "fake.example.com",
        "content": "fake",
        "ttl": 3600,
    },
    {
        "id": "fake2-id",
        "type": "TXT",
        "name": "fake2.example.com",
        "content": "fake2",
        "ttl": 3600,
    },
]


def assert_correct_output(capsys, expected_output_lines):
    out, _ = capsys.readouterr()
    assert out.splitlines() == expected_output_lines


def test_output_function_outputs_json_as_table(capsys):
    expected_output_lines = [
        "ID       TYPE NAME              CONTENT TTL ",
        "-------- ---- ----------------- ------- ----",
        "fake-id  TXT  fake.example.com  fake    3600",
        "fake2-id TXT  fake2.example.com fake2   3600",
    ]

    cli.handle_output(DATA, "TABLE", "list")
    assert_correct_output(capsys, expected_output_lines)


def test_output_function_outputs_json_as_table_with_no_header(capsys):
    expected_output_lines = [
        "fake-id  TXT fake.example.com  fake  3600",
        "fake2-id TXT fake2.example.com fake2 3600",
    ]

    cli.handle_output(DATA, "TABLE-NO-HEADER", "list")
    assert_correct_output(capsys, expected_output_lines)


def test_output_function_outputs_json_as_json_string(capsys):
    cli.handle_output(DATA, "JSON", "list")

    out, _ = capsys.readouterr()
    json_data = json.loads(out)

    assert json_data == DATA


def test_output_function_output_nothing_when_quiet(capsys):
    expected_output_lines = []

    cli.handle_output(DATA, "QUIET", "list")
    assert_correct_output(capsys, expected_output_lines)


def test_output_function_outputs_nothing_with_not_a_json_serializable(capsys):
    expected_output_lines = []

    cli.handle_output(object(), "TABLE", "list")
    assert_correct_output(capsys, expected_output_lines)

    cli.handle_output(object(), "TABLE-NO-HEADER", "list")
    assert_correct_output(capsys, expected_output_lines)

    cli.handle_output(object(), "JSON", "list")
    assert_correct_output(capsys, expected_output_lines)



"""Compatibility layer for Zeit (old name for Vercel)"""
import logging

from lexicon.providers.vercel import NAMESERVER_DOMAINS  # noqa: F401
from lexicon.providers.vercel import provider_parser  # noqa: F401
from lexicon.providers.vercel import Provider as VercelProvider

LOGGER = logging.getLogger(__name__)


class Provider(VercelProvider):
    """Provider for Zeit"""

    def __init__(self, config):
        LOGGER.error(
            "Zeit provider is deprecated and will be removed in a future version of Lexicon."
        )
        LOGGER.error("Please use Vercel provider instead.")
        super().__init__(config)



"""
To set enable live testing against the Exoscale API.

* create an account
* subscribe to DNS (with 1 domain)
* create a domain called "lexicontest.com"

Set the following environment variables.

    export LEXICON_EXOSCALE_KEY=EXO...
    export LEXICON_EXOSCALE_SECRET=xxx

"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
class ExoscaleProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for Exoscale"""

    provider_name = "exoscale"
    domain = "lexicontest.com"

    def _filter_headers(self):
        return ["X-DNS-Token", "x-request-id"]



"""Integration tests for Azure Cloud DNS"""
import re
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2

# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests


class AzureTests(TestCase, IntegrationTestsV2):
    """TestCase for Google Cloud DNS"""

    provider_name = "azure"
    domain = "full4ir.tk"

    def _test_parameters_overrides(self):
        return {"resource_group": "dns-test"}

    def _filter_headers(self):
        return [("Authorization", "Bearer TOKEN")]

    def _filter_post_data_parameters(self):
        return [("client_id", "CLIENT_ID"), ("client_secret", "CLIENT_SECRET")]

    def _filter_request(self, request):
        # Hide auth_tenant_id value in oauth token requests
        request.uri = re.sub(
            r"/[\w-]+/oauth2/token", "/TENANT_ID/oauth2/token", request.uri
        )
        # Hide auth_subscription_id value in DNS requests
        request.uri = re.sub(
            r"/subscriptions/[\w-]+/", "/subscriptions/SUBSCRIPTION_ID/", request.uri
        )
        return request

    def _filter_response(self, response):
        # Hide access_token value in oauth token responses
        response["body"]["string"] = re.sub(
            br'"access_token":"[\w.-]+"',
            b'"access_token":"TOKEN"',
            response["body"]["string"],
        )
        response["body"]["string"] = re.sub(
            br"\\/subscriptions\\/[\w-]+\\/",
            b"\\/subscriptions\\/SUBSCRIPTION_ID\\/",
            response["body"]["string"],
        )
        return response



"""Integration tests for DNSMadeEasy"""
from unittest import TestCase

import pytest

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
class DnsmadeeasyProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for DNSMadeEasy"""

    provider_name = "dnsmadeeasy"
    domain = "capsulecd.com"

    def _test_parameters_overrides(self):
        return {"api_endpoint": "http://api.sandbox.dnsmadeeasy.com/V2.0"}

    def _filter_headers(self):
        return ["x-dnsme-apiKey", "x-dnsme-hmac", "Authorization"]

    @pytest.mark.skip(reason="new test, missing recording")
    def test_provider_when_calling_update_record_should_modify_record_name_specified(
        self,
    ):
        return



"""Integration tests for EasyDNS"""
from unittest import TestCase

import pytest

from lexicon.tests.providers.integration_tests import IntegrationTestsV1


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
# TODO: migrate to IntegrationTestsV2 and its extended test suite
class EasyDnsProviderTests(TestCase, IntegrationTestsV1):
    """TestCase for EasyDNS"""

    provider_name = "easydns"
    domain = "easydnstemp.com"

    def _test_parameters_overrides(self):
        return {"api_endpoint": "http://sandbox.rest.easydns.net"}

    def _filter_headers(self):
        return ["Authorization"]

    def _filter_query_parameters(self):
        return ["_key", "_user"]

    @pytest.mark.skip(reason="new test, missing recording")
    def test_provider_when_calling_update_record_should_modify_record_name_specified(
        self,
    ):
        return



"""Integration tests for Onapp"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


class OnappProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for Oneapp"""

    provider_name = "onapp"
    domain = "my-test.org"

    def _filter_headers(self):
        return ["Authorization"]

    def _test_parameters_overrides(self):
        return {"auth_server": "https://dashboard.dynomesh.com.au"}



"""test module for aliyun provider"""

# Test for one implementation of the interface
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class AliyunProviderTests(TestCase, IntegrationTestsV2):
    """Integration tests for Foo provider"""

    provider_name = "aliyun"
    domain = "mean.space"

    def _filter_query_parameters(self):
        """filter access key id && secret info"""
        return [
            ("AccessKeyId", "DUMMY_KEY_ID"),
            ("Signature", "DUMMY_SIGNATURE"),
            ("SignatureNonce", "DUMMY_SIGNATURE_NONCE"),
            ("Timestamp", "DUMMY_TIMESTAMP"),
        ]



"""Integration tests for OVH"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class OvhProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for OVH"""

    provider_name = "ovh"
    domain = "pacalis.net"

    def _filter_headers(self):
        return ["X-Ovh-Application", "X-Ovh-Consumer", "X-Ovh-Signature"]

    def _test_parameters_overrides(self):
        return {"auth_entrypoint": "ovh-eu"}



"""Integration tests for Cloudflare"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
class CloudflareProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for Cloudflare"""

    provider_name = "cloudflare"
    domain = "pacalis.net"

    def _filter_headers(self):
        return ["X-Auth-Email", "X-Auth-Key", "set-cookie"]

    # We do not want to have "placeholder_auth_username" as default value for `--auth-username`
    # if Bearer tokens are used to execute the tests, because the non-emptiness of this flags
    # triggers interpretation of `--auth-key` as a Global API key.
    # Similarly for `--zone-id`, we want to control when its value is not empty, because
    # it will change the logic of the authentication process.
    def _test_fallback_fn(self):
        return (
            lambda x: "placeholder_" + x
            if x not in ("auth_username", "zone_id")
            else ""
        )



"""Integration tests for Dreamhost"""
from unittest import TestCase

import pytest

from lexicon.providers.dreamhost import Provider
from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class DreamhostProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for Dreamhost"""

    provider_name = "dreamhost"
    domain = "lexicon-example.com"

    def _filter_query_parameters(self):
        return ["key"]

    @pytest.mark.skip(reason="can not set ttl when creating/updating records")
    def test_provider_when_calling_list_records_after_setting_ttl(self):
        return

    def test_identifier(self):
        """Test _identifier"""
        dreamhost_record = {
            "type": "A",
            "record": "www.example.com",
            "value": "1.2.3.4",
        }
        identifier = Provider._identifier(dreamhost_record)

        dreamhost_record_from_id = Provider._id_to_dreamhost_record(identifier)

        assert dreamhost_record_from_id == dreamhost_record

    def test_id_to_record(self):
        """Test _id_to_record and _record_to_dreamhost_record"""
        dreamhost_record = {
            "type": "A",
            "record": "www.example.com",
            "value": "1.2.3.4",
        }

        identifier = Provider._identifier(dreamhost_record)

        record = Provider._id_to_record(identifier)
        dreamhost_record_from_id = Provider._record_to_dreamhost_record(record)

        assert dreamhost_record_from_id == dreamhost_record

    def test_id_to_dreamhost_record(self):
        """Test _id_to_dreamhost_record"""
        dreamhost_record = {
            "type": "A",
            "record": "www.example.com",
            "value": "1.2.3.4",
        }

        identifier = Provider._identifier(dreamhost_record)

        dreamhost_record_from_id = Provider._id_to_dreamhost_record(identifier)

        assert dreamhost_record_from_id == dreamhost_record



"""
Integration tests for EUserv provider

Author: Matthias Schoettle (@mschoettle), 2019
"""
import json
import re
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class EUservProviderTests(TestCase, IntegrationTestsV2):
    """Tests for EUserv"""

    # EUserv has a limit of 10 TXT entries. Therefore, the live recordings were
    # and entries removed manually during the execution of the tests.
    # Otherwise, a test case will fail with a corresponding error (maximum entries reached).

    provider_name = "euserv"
    domain = "schoettle.it"
    product_id_domain = 1

    def _filter_headers(self):
        return ["set-cookie"]

    def _filter_query_parameters(self):
        return [
            ("password", "PASSWORD"),
            ("email", "EMAIL"),
            ("sess_id", "SESSION_ID"),
            ("ord_no", "ORDER_ID"),
        ]

    def _filter_response(self, response):
        for cookie in ["set-cookie", "Set-Cookie"]:
            if cookie in response["headers"]:
                del response["headers"][cookie]

        if "string" in response["body"]:
            # Replace session and order id with placeholders
            response["body"]["string"] = re.sub(
                br'"sess_id":{"value":"[\w.-]+"',
                b'"sess_id":{"value":"SESSION_ID"',
                response["body"]["string"],
            )

            response["body"]["string"] = re.sub(
                br'"ord_no":{"value":"[\w.-]+"',
                b'"ord_no":{"value":"ORDER_ID"',
                response["body"]["string"],
            )

            # Replace orders in body with mock data (contains the minimal required data)
            filtered_body = json.loads(response["body"]["string"].decode("UTF-8"))

            if "result" in filtered_body and "orders" in filtered_body["result"]:
                filtered_body["result"]["orders"] = [
                    {
                        "ord_no": {"value": "ORDER_ID"},
                        "pg_id": {"value": self.product_id_domain},
                        "ord_description": {"value": "Contract Name\n" + self.domain},
                    }
                ]

            response["body"]["string"] = json.dumps(filtered_body).encode("UTF-8")

        return response



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lexicon', 'lexicon.providers', 'lexicon.tests', 'lexicon.tests.providers']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4>=4',
 'cryptography>=2',
 'pyyaml>=3',
 'requests>=2',
 'tldextract>=2']

extras_require = \
{'ddns': ['dnspython>=2'],
 'full': ['boto3>=1', 'localzone>=0.9.8', 'softlayer>=5', 'zeep>=3', 'oci>=2'],
 'gransy': ['zeep>=3'],
 'localzone': ['localzone>=0.9.8'],
 'oci': ['oci>=2'],
 'route53': ['boto3>=1'],
 'softlayer': ['softlayer>=5'],
 'transip': ['transip>=2']}

entry_points = \
{'console_scripts': ['lexicon = lexicon.cli:main']}

setup_kwargs = {
    'name': 'dns-lexicon',
    'version': '3.8.3',
    'description': 'Manipulate DNS records on various DNS providers in a standardized/agnostic way',
    'long_description': '============\n|logo_named|\n============\n\nManipulate DNS records on various DNS providers in a standardized/agnostic way.\n\n|build_status| |coverage_status| |docker_pulls| |pypy_version| |pypy_python_support| |github_license|\n\n.. |logo_named| image:: https://raw.githubusercontent.com/AnalogJ/lexicon/master/docs/images/logo_named.svg\n    :alt: Lexicon\n\n.. |build_status| image:: https://dev.azure.com/AnalogJ/lexicon/_apis/build/status/AnalogJ.lexicon?branchName=master\n    :target: https://dev.azure.com/AnalogJ/lexicon/_build/latest?definitionId=1&branchName=master\n\n.. |coverage_status| image:: https://coveralls.io/repos/github/AnalogJ/lexicon/badge.svg\n    :target: https://coveralls.io/github/AnalogJ/lexicon?branch=master\n\n.. |docker_pulls| image:: https://img.shields.io/docker/pulls/analogj/lexicon.svg\n    :target: https://hub.docker.com/r/analogj/lexicon\n\n.. |pypy_version| image:: https://img.shields.io/pypi/v/dns-lexicon.svg\n    :target: https://pypi.python.org/pypi/dns-lexicon\n\n.. |pypy_python_support| image:: https://img.shields.io/pypi/pyversions/dns-lexicon.svg\n    :target: https://pypi.python.org/pypi/dns-lexicon\n\n.. |github_license| image:: https://img.shields.io/github/license/AnalogJ/lexicon.svg\n    :target: https://github.com/AnalogJ/lexicon/blob/master/LICENSE\n\n.. contents:: Table of Contents\n   :local:\n\n.. tag:intro-begin\n\nWhy using Lexicon?\n==================\n\nLexicon provides a way to manipulate DNS records on multiple DNS providers in a standardized way.\n\nLexicon can be used as:\n\n- a CLI tool:\n\n.. code-block:: bash\n\n    # Create a TXT entry in domain.net zone hosted by CloudFlare\n    lexicon cloudflare create domain.net TXT --name foo --content bar\n\n- or a Python library:\n\n.. code-block:: python\n\n    # Create a TXT entry in domain.net zone hosted by CloudFlare\n    from lexicon.client import Client\n    from lexicon.config import ConfigResolver\n\n    action = {\n        "provider_name" : "cloudflare",\n        "action": "create",\n        "domain": "domain.net",\n        "type": "TXT",\n        "name": "foo",\n        "content": "bar",\n    }\n    config = ConfigResolver().with_env().with_dict(action)\n    Client(config).execute()\n\nLexicon was designed to be used in automation, specifically letsencrypt.\n\n* `Generating Intranet & Private Network SSL Certificates using Lets Encrypt & Lexicon <http://blog.thesparktree.com/post/138999997429/generating-intranet-and-private-network-ssl>`_\n\nSupported providers\n===================\n\nOnly DNS providers who have an API can be supported by `lexicon`.\n\nThe current supported providers are:\n\n- `Aliyun.com <https://help.aliyun.com/document_detail/29739.html>`_\n- `AuroraDNS <https://www.pcextreme.com/aurora/dns>`_\n- `AWS Route53 <https://docs.aws.amazon.com/Route53/latest/APIReference/Welcome.html>`_\n- `Azure DNS <https://docs.microsoft.com/en-us/rest/api/dns/>`_\n- `Cloudflare <https://api.cloudflare.com/#endpoints>`_\n- `ClouDNS <https://www.cloudns.net/wiki/article/56/>`_\n- `CloudXNS <https://www.cloudxns.net/Support/lists/cid/17.html>`_\n- `ConoHa <https://www.conoha.jp/docs/>`_\n- `Constellix <https://api-docs.constellix.com/?version=latest>`_\n- `DigitalOcean <https://developers.digitalocean.com/documentation/v2/#create-a-new-domain>`_\n- `Dinahosting <https://en.dinahosting.com/api>`_\n- `DirectAdmin <https://www.directadmin.com/features.php?id=504>`_\n- DNSimple `v1 <https://developer.dnsimple.com/>`_, `v2 <https://developer.dnsimple.com/v2/>`_\n- `DnsMadeEasy <https://api-docs.dnsmadeeasy.com/?version=latest>`_\n- `DNSPark <https://dnspark.zendesk.com/entries/31210577-REST-API-DNS-Documentation>`_\n- `DNSPod <https://support.dnspod.cn/Support/api>`_\n- `Dreamhost <https://help.dreamhost.com/hc/en-us/articles/217560167-API_overview>`_\n- `Dynu <https://www.dynu.com/Support/API>`_\n- `EasyDNS <http://docs.sandbox.rest.easydns.net/>`_\n- `Easyname <https://www.easyname.com/en>`_\n- `EUserv <https://support.euserv.com/api-doc/>`_\n- `ExoScale <https://community.exoscale.com/documentation/dns/api/>`_\n- Gandi `RPC (old) <http://doc.rpc.gandi.net>`_ / `LiveAPI <http://doc.livedns.gandi.net/>`_\n- `Gehirn <https://support.gehirn.jp/apidocs/gis/dns/index.html>`_\n- `Glesys <https://github.com/glesys/API/wiki/>`_\n- `GoDaddy <https://developer.godaddy.com/getstarted#access>`_\n- `Google Cloud DNS <https://cloud.google.com/dns/api/v1/>`_\n- `Gransy (sites subreg.cz, regtons.com and regnames.eu) <https://subreg.cz/manual/>`_\n- `Hover <https://hoverapi.docs.apiary.io/>`_\n- `Hurricane Electric DNS <https://dns.he.net/>`_\n- `Hetzner <https://dns.hetzner.com/api-docs/>`_\n- `Infoblox <https://docs.infoblox.com/display/ILP/Infoblox+Documentation+Portal>`_\n- `Infomaniak <https://www.infomaniak.com>`_\n- `Internet.bs <https://internetbs.net/ResellerRegistrarDomainNameAPI>`_\n- `INWX <https://www.inwx.de/en/offer/api>`_\n- `Joker.com <https://joker.com/faq/index.php?action=show&cat=39>`_\n- `Linode <https://www.linode.com/api/dns>`_\n- `Linode v4 <https://developers.linode.com/api/docs/v4#tag/Domains>`_\n- `LuaDNS <http://www.luadns.com/api.html>`_\n- `Memset <https://www.memset.com/apidocs/methods_dns.html>`_\n- `Mythic Beasts (v2 API) <https://www.mythic-beasts.com/support/api/dnsv2>`_\n- `Njalla <https://njal.la/api/>`_\n- `Namecheap <https://www.namecheap.com/support/api/methods.aspx>`_\n- `Namesilo <https://www.namesilo.com/api_reference.php>`_\n- `Netcup <https://ccp.netcup.net/run/webservice/servers/endpoint.php>`_\n- NFSN (NearlyFreeSpeech)\n- `NS1 <https://ns1.com/api/>`_\n- `OnApp <https://docs.onapp.com/display/55API/OnApp+5.5+API+Guide>`_\n- Online\n- `OVH <https://api.ovh.com/>`_\n- `Plesk <https://docs.plesk.com/en-US/onyx/api-rpc/about-xml-api.28709/>`_\n- `PointHQ <https://pointhq.com/api/docs>`_\n- `PowerDNS <https://doc.powerdns.com/md/httpapi/api_spec/>`_\n- `Rackspace <https://developer.rackspace.com/docs/cloud-dns/v1/developer-guide/>`_\n- `Rage4 <https://gbshouse.uservoice.com/knowledgebase/articles/109834-rage4-dns-developers-api>`_\n- `RcodeZero <https://my.rcodezero.at/api-doc>`_\n- `RFC2136 <https://en.wikipedia.org/wiki/Dynamic_DNS>`_\n- `Sakura Cloud by SAKURA Internet Inc. <https://developer.sakura.ad.jp/cloud/api/1.1/>`_\n- `SafeDNS by UKFast <https://developers.ukfast.io/documentation/safedns>`_\n- `SoftLayer <https://sldn.softlayer.com/article/REST#HTTP_Request_Types>`_\n- `Transip <https://www.transip.nl/transip/api/>`_\n- `UltraDNS <https://ultra-portalstatic.ultradns.com/static/docs/REST-API_User_Guide.pdf>`_\n- `Vercel <https://vercel.com/docs/api#endpoints/dns>`_\n- `Vultr <https://www.vultr.com/api/#tag/dns>`_\n- `Yandex <https://tech.yandex.com/domain/doc/reference/dns-add-docpage/>`_\n- `Zilore <https://zilore.com/en/help/api>`_\n- `Zonomi <http://zonomi.com/app/dns/dyndns.jsp>`_\n\n.. tag:intro-end\n\nDocumentation\n=============\n\nOnline documentation (user guide, configuration reference) is available in the `Lexicon documentation`_.\n\nFor a quick start, please have a look in particular at the `User guide`_.\n\n.. _Lexicon documentation: https://dns-lexicon.readthedocs.io\n.. _User guide: https://dns-lexicon.readthedocs.io/en/latest/user_guide.html\n\nContributing\n============\n\nIf you want to help in the Lexicon development, you are welcome!\n\nPlease have a look at the `Developer guide`_ page to know how to start.\n\n.. _Developer guide: https://dns-lexicon.readthedocs.io/en/latest/developer_guide.html\n\nLicensing\n=========\n\n- MIT\n- Logo_: transform by Mike Rowe from the Noun Project\n\n.. _Logo: https://thenounproject.com/term/transform/397964\n',
    'author': 'Jason Kulatunga',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/AnalogJ/lexicon',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""Integration tests for Hover"""
import json
import re
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


class HoverProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for Hover"""

    provider_name = "hover"
    domain = "novuslex.com"
    domain_id = "dom1127777"
    hoverauth = "0123456789abcdef0123456789abcdef"
    hover_session = "0123456789abcdef0123456789abcdef"

    def _filter_post_data_parameters(self):
        return ["username", "password"]

    def _filter_headers(self):
        return ["Cookie"]

    def _filter_query_parameters(self):
        return ["hover_session", "hoverauth"]

    def _replace_auth(self, cookie):
        cookie = re.sub(
            "hover_session=.*;", f"hover_session={self.hover_session};", cookie
        )
        cookie = re.sub("hoverauth=.*;", f"hoverauth={self.hoverauth};", cookie)
        return cookie

    def _filter_response(self, response):
        if "basestring" not in globals():
            basestring = str

        if "set-cookie" in response["headers"]:
            if isinstance(response["headers"]["set-cookie"], basestring):
                response["headers"]["set-cookie"] = self._replace_auth(
                    response["headers"]["set-cookie"]
                )
            else:
                for i, cookie in enumerate(response["headers"]["set-cookie"]):
                    response["headers"]["set-cookie"][i] = self._replace_auth(cookie)

        try:
            filtered_body = json.loads(response["body"]["string"].decode("UTF-8"))
        except ValueError:
            # Body is not json during authentication, so we let it through.
            # Helper function _request in hover.py will raise exception when
            # response is not json and it should be.
            pass
        else:
            # filter out my personal contact information
            if "contact" in filtered_body:
                del filtered_body["contact"]

            # if the response is listing all my domains then return a mock response
            if "domains" in filtered_body and len(filtered_body["domains"]) > 1:
                filtered_body["domains"] = [
                    {
                        "id": self.domain_id,
                        "domain_name": self.domain,
                        "status": "active",
                    }
                ]

            response["body"]["string"] = json.dumps(filtered_body).encode("UTF-8")
        return response



"""Integration tests for Softlayer"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV1


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
# TODO: migrate to IntegrationTestsV2 and its extended test suite
class SoftLayerProviderTests(TestCase, IntegrationTestsV1):
    """TestCase for SoftLayer"""

    provider_name = "softlayer"
    domain = "example.com"

    # SoftLayer does not provide a sandbox API; actual credentials are required
    # Keeping this here for when fixtures need to be regenerated
    # def _test_options(self):
    #    options = super(SoftLayerProviderTests, self)._test_options()
    #    options.update({
    #        'auth_username': 'foo',
    #        'auth_api_key': 'bar'
    #        })
    #    return options



"""Integration tests for HostingDE provider"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class FooProviderTests(TestCase, IntegrationTestsV2):
    """Integration tests for hostingde provider"""

    provider_name = "hostingde"
    domain = "eruza.de"

    def _filter_post_data_parameters(self):
        return ["authToken"]

    def _filter_headers(self):
        return ["Authorization"]

    def _filter_query_parameters(self):
        return ["secret_key"]



"""Integration tests for Dynu.com"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class DynuProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for Dynu.com"""

    provider_name = "dynu"
    domain = "example.com"

    def _filter_headers(self):
        return ["API-Key"]

    def _filter_response(self, response):
        if response["status"]["code"] not in [200, 503]:
            return None
        return response



"""Integration tests for Njalla provider"""
from unittest import TestCase

import pytest

from lexicon.tests.providers.integration_tests import IntegrationTestsV2

# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests


class NjallaProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for Njalla"""

    provider_name = "njalla"
    domain = "example.com"

    def _filter_headers(self):
        return ["Authorization"]

    @pytest.mark.skip(reason="provider allows duplicate records")
    def test_provider_when_calling_create_record_with_duplicate_records_should_be_noop(
        self,
    ):
        return

    @pytest.mark.skip(reason="provider does not recognize record sets")
    def test_provider_when_calling_delete_record_with_record_set_name_remove_all(self):
        return



"""Integration tests for Plesk"""
from unittest import TestCase

import pytest

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class PleskProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for Plesk"""

    provider_name = "plesk"
    domain = "lexicon-test.com"

    def _filter_headers(self):
        return ["HTTP_AUTH_LOGIN", "HTTP_AUTH_PASSWD"]

    def _test_parameters_overrides(self):
        return {"plesk_server": "https://quasispace.de:8443"}

    @pytest.mark.skip(reason="can not set ttl when creating/updating records")
    def test_provider_when_calling_list_records_after_setting_ttl(self):
        return



"""Integration tests for Localzone"""
import os
from unittest import TestCase

import pytest

from lexicon.tests.providers.integration_tests import IntegrationTestsV2

FILENAME = "db.example.com"
ZONEFILE = """
$ORIGIN example.com.     ; designates the start of this zone file in the namespace
$TTL 1h                  ; default expiration time of all resource records without their own TTL value
example.com.  IN  SOA   ns.example.com. username.example.com. ( 2007120710 1d 2h 4w 1h )
example.com.  IN  NS    ns                    ; ns.example.com is a nameserver for example.com
example.com.  IN  NS    ns.somewhere.example. ; ns.somewhere.example is a backup nameserver for example.com
example.com.  IN  MX    10 mail.example.com.  ; mail.example.com is the mailserver for example.com
@             IN  MX    20 mail2.example.com. ; equivalent to above line, "@" represents zone origin
@             IN  MX    50 mail3              ; equivalent to above line, but using a relative host name
example.com.  IN  A     192.0.2.1             ; IPv4 address for example.com
              IN  AAAA  2001:db8:10::1        ; IPv6 address for example.com
ns            IN  A     192.0.2.2             ; IPv4 address for ns.example.com
              IN  AAAA  2001:db8:10::2        ; IPv6 address for ns.example.com
www           IN  CNAME example.com.          ; www.example.com is an alias for example.com
wwwtest       IN  CNAME www                   ; wwwtest.example.com is another alias for www.example.com
mail          IN  A     192.0.2.3             ; IPv4 address for mail.example.com
mail2         IN  A     192.0.2.4             ; IPv4 address for mail2.example.com
mail3         IN  A     192.0.2.5             ; IPv4 address for mail3.example.com
@             IN  TXT   "v=spf1 mx ~all"      ; SPFv1 record for example.com
"""


@pytest.fixture(scope="module", autouse=True)
def testfile():
    """Create a local zone file for testing."""
    test_file = open(FILENAME, "w")
    test_file.write(ZONEFILE)
    test_file.close()
    yield
    os.remove(FILENAME)


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
class LocalzoneProviderTests(TestCase, IntegrationTestsV2):
    """Integration tests for Localzone"""

    provider_name = "localzone"
    domain = "example.com"

    def _test_parameters_overrides(self):
        options = {"filename": FILENAME}

        return options

    def _test_fallback_fn(self):
        return lambda _: None

    @pytest.mark.skip(reason="localzone does not require authentication")
    def test_provider_authenticate(self):
        return

    @pytest.mark.skip(reason="localzone does not require authentication")
    def test_provider_authenticate_with_unmanaged_domain_should_fail(self):
        return



"""Module provider for Dynu.com"""
import json
import logging

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["dynu.com"]
FREE_SERVICE_LIMIT = 4
MEMBERSHIP_DETAILS = "https://www.dynu.com/en-US/Membership"


def provider_parser(subparser):
    """Module provider for Dynu.com"""
    subparser.add_argument("--auth-token", help="specify api key for authentication")


class Provider(BaseProvider):
    """Provider class for Dynu.com"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://api.dynu.com/v2"

    def _authenticate(self):
        data = self._get("/dns")
        domains = data["domains"]
        for domain in domains:
            if domain["name"].lower() == self.domain.lower():
                self.domain_id = domain["id"]
                break
        else:
            raise AuthenticationError("No matching domain found")

    # Create record. If record already exists with the same content, do nothing.
    def _create_record(self, rtype, name, content):
        record = self._to_dynu_record(rtype, name, content)

        if self._get_lexicon_option("ttl"):
            record["ttl"] = self._get_lexicon_option("ttl")

        created = False
        try:
            payload = self._post(f"/dns/{self.domain_id}/record", record)
            created = self._from_dynu_record(payload)
        except requests.exceptions.HTTPError as error:
            # HTTP 501 is expected when a record with the same type and content is sent to the
            # server.
            if error.response.status_code == 501:
                created = True
            else:
                raise error
        LOGGER.debug("create_record: %s", created)
        return created

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        payload = self._get(f"/dns/{self.domain_id}/record")

        records = []
        for record in payload["dnsRecords"]:
            processed_record = self._from_dynu_record(record)
            records.append(processed_record)

        len_all = len(records)

        if rtype:
            records = [record for record in records if record["type"] == rtype]

        if name:
            records = [
                record for record in records if record["name"] == self._full_name(name)
            ]

        if content:
            records = [record for record in records if record["content"] == content]

        len_removed = len_all - len(records)
        if len_removed:
            LOGGER.debug("list_records: removed %d, total %d", len_removed, len_all)

        LOGGER.debug("list_records: %s", records)
        return records

    # Update a record.
    def _update_record(self, identifier, rtype=None, name=None, content=None):
        records = {}
        if identifier is None:
            records = self._list_records(rtype, name, None)
            records = {rec["id"]: rec for rec in records}
        else:
            if name is not None:
                # check if new name matches old one
                # dynu prohibits changing the node name, so we have to delete
                # and create a new record then
                original = self._fetch_record(identifier)
                orig_name = self._relative_name(original["name"]).lower()
                new_name = self._relative_name(name).lower()
                if orig_name != new_name:
                    LOGGER.info(
                        "Cannot change node name from %s to %s, deleting %s and creating new",
                        orig_name,
                        new_name,
                        identifier,
                    )
                    self._delete_record(identifier)
                    return self._create_record(rtype, name, content)

            records = {identifier: self._to_dynu_record(rtype, None, content)}

        for ident, rec in records.items():
            add_rec = self._to_dynu_record(
                rec["type"], self._relative_name(rec["name"]), rec["content"]
            )
            payload = self._post(f"/dns/{self.domain_id}/record/{ident}", add_rec)
            update = self._from_dynu_record(payload)
            LOGGER.debug("update_record: %s", update)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self._list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)

        for record_id in delete_record_id:
            try:
                self._delete(f"/dns/{self.domain_id}/record/{record_id}")
                LOGGER.debug("delete_record: %s", record_id)
            except requests.exceptions.HTTPError as error:
                if error.response.status_code == 501:
                    LOGGER.info("delete_record: %s does not exist", record_id)
                    continue
                raise error

        return True

    # Helpers
    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data:
            data = json.dumps(data)

        LOGGER.debug("Request: %s %s with data %s", action, url, data)

        response = requests.request(
            action,
            self.api_endpoint + url,
            data=data,
            headers={
                "API-Key": self._get_provider_option("auth_token"),
                "Accept": "application/json",
                "Content-Type": "application/json",
            },
        )
        LOGGER.debug("Response: %s", response)
        # if the request fails for any reason, throw an error.
        if response.status_code == 503:
            raise RuntimeError(
                "Too many entries, limit for free service is {0}, see {1}".format(
                    FREE_SERVICE_LIMIT, MEMBERSHIP_DETAILS
                )
            )
        response.raise_for_status()
        return response.json()

    # Fetch a record by its ID
    def _fetch_record(self, identifier):
        payload = self._get(f"/dns/{self.domain_id}/record/{identifier}")
        return self._from_dynu_record(payload)

    # Takes a Dynu.com record and puts it into lexicon-shape
    @staticmethod
    def _from_dynu_record(record):
        rtype = record["recordType"]
        options = {"enabled": "state", "lastUpdate": "updatedOn"}
        options = {k: record[v] for k, v in options.items() if v in record}
        options["raw"] = record

        # map additional fields depending on the record type
        # the result takes the record type as key, and all options of the
        # matching key in the following dict as dict of values,
        # e.g.
        # 'options': {
        #     'CNAME': {
        #         'host': 'example.com'
        #     }
        # }
        mapping = {
            "A": {"ipv4": "ipv4Address", "group": "group"},
            "AAAA": {"ipv6": "ipv6Address", "group": "group"},
            "CNAME": {"host": "host"},
            "LOC": {
                "lat": "latitude",
                "long": "longitude",
                "alt": "altitude",
                "size": "size",
                "hPrec": "horizontalPrecision",
                "vPrec": "verticalPrecision",
            },
            "MX": {"host": "host", "priority": "priority"},
            "NS": {"host": "host"},
            "PTR": {"host": "host"},
            "SRV": {"host": "host", "priority": "priority", "weight": "weight"},
            "TXT": {"data": "textData"},
        }.get(rtype, {})
        options[rtype] = {k: record[v] for k, v in mapping.items() if v in record}

        out_record = {
            "id": record["id"],
            "type": rtype,
            "name": record["hostname"],
            "ttl": record["ttl"],
            "options": options,
        }

        # format the content as noted in the spec, e.g. take everything
        # in the raw DNS response after the record type, and remove quotations
        # Example:
        #   example.com. 120 IN TXT \"txt-value=thisIsATest\"
        # Becomes:
        #   txt-value=thisIsATest
        if record["content"]:
            content = record["content"]
            content = content.split(rtype)[1]
            content = content.strip()
            content = content.replace('"', "")
            out_record["content"] = content

        return out_record

    # Takes record input and puts it into a format the Dynu.com API supports
    def _to_dynu_record(self, rtype, name, content):
        if rtype == "LOC":
            raise NotImplementedError("LOC is not available for this provider")

        if rtype == "SPF":
            raise NotImplementedError("SPF entries are not supported")

        output = {"recordType": rtype, "state": True}
        if name is not None:
            output["nodeName"] = self._relative_name(name)

        cnt_split = content.split(" ")
        output.update(
            {
                "A": (lambda: {"ipv4Address": content}),
                "AAAA": (lambda: {"ipv6Address": content}),
                "CNAME": (lambda: {"host": content}),
                "MX": (lambda: {"priority": cnt_split[0], "host": cnt_split[1]}),
                "NS": (lambda: {"host": content}),
                "PTR": (lambda: {"host": content}),
                "SRV": (
                    lambda: {
                        "priority": cnt_split[0],
                        "weight": cnt_split[1],
                        "port": cnt_split[2],
                        "host": cnt_split[3],
                    }
                ),
                "TXT": (lambda: {"textData": content}),
            }.get(rtype, lambda: {})()
        )
        return output



"""Integration tests for GratisDNS"""
import re
from unittest import TestCase

from lexicon.providers.gratisdns import Provider
from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class GratisDNSProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for GratisDNS"""

    Provider = Provider
    provider_name = "gratisdns"
    domain = "denisa.dk"
    gratisdns_session = "0123456789abcdef0123456789abcdef"

    def _filter_post_data_parameters(self):
        return ["login", "password"]

    def _filter_headers(self):
        return ["Cookie"]

    def _replace_auth(self, cookie):
        cookie = re.sub("ORGID=.*;", f"ORGID={self.gratisdns_session};", cookie)
        return cookie

    # Inspired by thehover provider
    def _filter_response(self, response):
        if "basestring" not in globals():
            basestring = str

        if "set-cookie" in response["headers"]:
            if isinstance(response["headers"]["set-cookie"], basestring):
                response["headers"]["set-cookie"] = self._replace_auth(
                    response["headers"]["set-cookie"]
                )
            else:
                for i, cookie in enumerate(response["headers"]["set-cookie"]):
                    response["headers"]["set-cookie"][i] = self._replace_auth(cookie)

        return response



"""Integration tests for Infomaniak"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class InfomaniakProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for Infomaniak"""

    provider_name = "infomaniak"
    domain = "testouille.at"

    def _filter_headers(self):
        return ["Authorization"]



"""Provider module for CloudNS"""
import logging

import requests

from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["cloudns.net"]


def provider_parser(subparser):
    """Configure provider parser for CloudNS"""
    identity_group = subparser.add_mutually_exclusive_group()
    identity_group.add_argument("--auth-id", help="specify user id for authentication")
    identity_group.add_argument(
        "--auth-subid", help="specify subuser id for authentication"
    )
    identity_group.add_argument(
        "--auth-subuser", help="specify subuser name for authentication"
    )
    subparser.add_argument(
        "--auth-password", help="specify password for authentication"
    )
    subparser.add_argument("--weight", help="specify the SRV record weight")
    subparser.add_argument("--port", help="specify the SRV record port")


class Provider(BaseProvider):
    """Provider class for CloudNS"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://api.cloudns.net"

    def _authenticate(self):
        payload = self._get("/dns/get-zone-info.json", {"domain-name": self.domain})
        self.domain_id = payload["name"]
        LOGGER.debug("authenticate: %s", payload)

    def _create_record(self, rtype, name, content):
        # Skip execution if such a record already exists
        existing_records = self._list_records(rtype, name, content)
        if existing_records:
            return True

        # Build parameters for adding a new record
        params = {
            "domain-name": self.domain_id,
            "record-type": rtype,
            "host": self._relative_name(name),
            "record": content,
        }
        if self._get_lexicon_option("ttl"):
            params["ttl"] = self._get_lexicon_option("ttl")
        if self._get_lexicon_option("priority"):
            params["priority"] = self._get_lexicon_option("priority")
        if self._get_provider_option("weight"):
            params["weight"] = self._get_lexicon_option("weight")
        if self._get_provider_option("port"):
            params["port"] = self._get_lexicon_option("port")

        # Add new record by calling the ClouDNS API
        payload = self._post("/dns/add-record.json", params)
        LOGGER.debug("create_record: %s", payload)

        # Error handling is already covered by self._request
        return True

    def _list_records(self, rtype=None, name=None, content=None):
        # Build parameters to make use of the built-in API filtering
        params = {"domain-name": self.domain_id}
        if rtype:
            params["type"] = rtype
        if name:
            params["host"] = self._relative_name(name)

        # Fetch and parse all records for the given zone
        payload = self._get("/dns/records.json", params)
        payload = payload if not isinstance(payload, list) else {}
        records = []
        for record in payload.values():
            records.append(
                {
                    "type": record["type"],
                    "name": self._full_name(record["host"]),
                    "ttl": record["ttl"],
                    "content": record["record"],
                    "id": record["id"],
                }
            )

        # Filter by content manually as API does not support that
        if content:
            records = [record for record in records if record["content"] == content]

        # Print records as debug output and return them
        LOGGER.debug("list_records: %s", records)
        return records

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        # Try to find record if no identifier was specified
        if not identifier:
            identifier = self._find_record_identifier(rtype, name, None)

        # Build parameters for updating an existing record
        params = {"domain-name": self.domain_id, "record-id": identifier}
        if name:
            params["host"] = self._relative_name(name)
        if content:
            params["record"] = content
        if self._get_lexicon_option("ttl"):
            params["ttl"] = self._get_lexicon_option("ttl")
        if self._get_lexicon_option("priority"):
            params["priority"] = self._get_lexicon_option("priority")
        if self._get_provider_option("weight"):
            params["weight"] = self._get_provider_option("weight")
        if self._get_provider_option("port"):
            params["port"] = self._get_provider_option("port")

        # Update existing record by calling the ClouDNS API
        payload = self._post("/dns/mod-record.json", params)
        LOGGER.debug("update_record: %s", payload)

        # Error handling is already covered by self._request
        return True

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        # Try to find record if no identifier was specified
        delete_record_id = []
        if not identifier:
            records = self._list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)

        for record_id in delete_record_id:
            # Delete existing record by calling the ClouDNS API
            self._post(
                "/dns/delete-record.json",
                {"domain-name": self.domain_id, "record-id": record_id},
            )

        LOGGER.debug("delete_record: %s", True)

        # Error handling is already covered by self._request
        return True

    def _build_authentication_data(self):
        if not self._get_provider_option("auth_password"):
            raise Exception(
                "No valid authentication data passed, expected: auth-password"
            )

        if self._get_provider_option("auth_id"):
            return {
                "auth-id": self._get_provider_option("auth_id"),
                "auth-password": self._get_provider_option("auth_password"),
            }
        if self._get_provider_option("auth_subid"):
            return {
                "sub-auth-id": self._get_provider_option("auth_subid"),
                "auth-password": self._get_provider_option("auth_password"),
            }
        if self._get_provider_option("auth_subuser"):
            return {
                "sub-auth-user": self._get_provider_option("auth_subuser"),
                "auth-password": self._get_provider_option("auth_password"),
            }
        if (
            self._get_provider_option("auth_id")
            or self._get_provider_option("auth_subid")
            or self._get_provider_option("auth_subuser")
        ):
            # All the options were passed with a fallback value, return an empty dictionary.
            return {}
        raise Exception(
            "No valid authentication data passed, expected: auth-id, auth-subid, auth-subuser"
        )

    def _find_record_identifier(self, rtype, name, content):
        records = self._list_records(rtype, name, content)
        LOGGER.debug("records: %s", records)
        if len(records) == 1:
            return records[0]["id"]
        raise Exception("Record identifier could not be found.")

    def _request(self, action="GET", url="/", data=None, query_params=None):
        # Set default values for missing arguments
        data = data if data else {}
        query_params = query_params if query_params else {}

        # Merge authentication data into request
        if action == "GET":
            query_params.update(self._build_authentication_data())
        else:
            data.update(self._build_authentication_data())

        # Fire request against ClouDNS API and parse result as JSON
        response = requests.request(
            action, self.api_endpoint + url, params=query_params, data=data
        )
        response.raise_for_status()
        payload = response.json()

        # Check ClouDNS specific status code and description
        if (
            "status" in payload
            and "statusDescription" in payload
            and payload["status"] != "Success"
        ):
            raise Exception(
                "ClouDNS API request has failed: " + payload["statusDescription"]
            )

        # Return payload
        return payload
